'
' datafolder.bi
'
' FreeBASIC library for "data folders" - normally compressed folders containing application data
' allows to access an uncompressed folder as an identically named compressed archive file
' for example, a folder might be named "myproject.rsc", and all of its content will be
' packed into a ZIP-file that is renamed "myproject.rsc". Using the datafolder library,
' you can treat both identically.
'
' The library determines automatically whether you are having the uncompressed folder,
' or a compressed archive, and determines the type of archive/compression automatically.
' Further, normal files will be loaded just like using BLOAD, so libraries supporting
' data folders can also access normal files
'
' You can't change/edit the data files, resp. only with outside tools, but the
' datafolder library is only intended for static program data anyway.
'
' IMPORTANT: "filepath" is consisting of 2 parts:
'               - the file system path locating the data folder
'               - the internal data folder file path addressing a file inside the
'                 data folder
'               those are divided with a colon (":"), the path may contain e.g. a "A:" or
'               "http://", only the last colon in the path is considered as divider
'
' DFEXISTS(STRING filepath)
'        returns whether specified file is existing (inside data folder)
'        use "<path>/<datafolder>:" to check, whether datafolder exists
'
' DFISAFOLDER(STRING filepath)
'        returns whether specified datafolder is uncompressed(TRUE), i.e. an ordiary
'        folder, not an archive file
'        filepath may be datafolder or a file inside datafolder - with ":", or the path
'        of the datafolder, with and without ":". => Specified files may even not exist.
'
' DFLOAD(STRING filepath, STRING dest_buffer, STRING encryption="") loads the specified
'        data file into the the given destination string, automatically sizing
'
'        encryption is "<algorithm>-><key>", optionally for encrypted files or archives
'        "algorithm" is, for instance:
'          - 'XOR': super-primitive XOR encryption, "key" is a string that is repeatetly
'            XORed with the file content, just makes the content unreadable
'          - 'ZIP': PKZIP "ZIP-Crypto", used by ZIP archives
'          - 'AES': AES-128, AES-192 and AES-256, determined by the key length
'        If the key is contained somewhere in the application package, it won't be safe anyway,
'        just hindering ordinary users to read e.g. texts they shouldn't read (mostly spoilers
'        in computer games' data).
'        Asymmetric encryption might avoid manipulation of application data (also mostly cheating
'        in computer games).
'
' DFLOAD(STRING filepath, BYTE PTR dest_buffer, STRING encryption="")
'        alternate function working with an ALLOCATE-created buffer (demands a DEALLOCATE call)
'
' DFLOADIMAGE(STRING filepath, FB.IMAGE PTR dest_buffer, STRING encryption=""[, ULONG PTR palette])
'        loads an image file like BLOAD, dest_buffer must be deleted with IMAGEDESTROY after use
'        supported image formats: - as supported; at least BMP, propably PNG, JPG, GIF
'        automatically determines format and sizes, takes the bit depth of the file
'        palette has to be a ULONG[256] buffer if used, each entry is &h00RRGGBB value
'
' Future plan: DFDIR(pattern) to list the contents of a subdirectory

#INCLUDE ONCE "file.bi"
#INCLUDE ONCE "fbext.bi"
#INCLUDE ONCE "fbgfx.bi"
#INCLUDE ONCE "FBimage.bi"
#INCLUDE ONCE "system.bi"

FUNCTION DFISAFOLDER(BYVAL file AS STRING) AS BOOLEAN
  DIM AS INTEGER i
  DIM AS STRING filepath

  filepath=GETFILEPATH(file)

  i=INSTRREV(filepath, ":")
  IF i THEN
    i-=1
    filepath=LEFT(filepath, i)
  ENDIF

  DIR(filepath, &HFF, @i)
  IF (i AND &H10) THEN RETURN TRUE ' file type is folder

  ' #note: returns false, also if file/folder not existing => use DFEXISTS() !
  RETURN FALSE
END FUNCTION


FUNCTION DFEXISTS(BYVAL file AS STRING) AS BOOLEAN
  DIM AS INTEGER i
  DIM AS STRING archive_name
  DIM AS STRING filepath

  filepath=GETFILEPATH(file)

  i=INSTRREV(filepath, ":")
  IF i THEN
    IF DFISAFOLDER(filepath) THEN ' uncompressed data folder
      ' => check if file exists
      IF LEN(filepath)=i THEN ' last character is ":"
        ' => only checking whether datafolder exists
        i-=1
        filepath=MID(filepath, 1, i)
        ' #note: "attribute" &hFF should take every kind of file including folders,
        '        but this is not the recommended usage of DIR() and might not work
        '        always
        IF DIR(filepath, &hFF)>"" THEN RETURN TRUE
      ELSE
        i-=1
        filepath[i]=47 ' replace : with /
        IF FILEEXISTS(filepath) THEN RETURN TRUE
      ENDIF
    ELSE ' datafolder is a compressed archive
      archive_name=LEFT(filepath,i)
      filepath=MID(filepath, i+1)
      ' #todo: determine archive format, open and check for the file...
    ENDIF
  ENDIF

  ' any fail means "doesn't exist"...
  RETURN FALSE
END FUNCTION


DECLARE FUNCTION DFLOAD OVERLOAD (BYVAL file AS STRING, BYREF fbuffer AS STRING, BYVAL encryption AS STRING="") AS BOOLEAN
FUNCTION DFLOAD(BYVAL file AS STRING, BYREF fbuffer AS STRING, BYVAL encryption AS STRING="") AS BOOLEAN
  DIM AS INTEGER i, l, fn
  DIM AS STRING archive_name
  DIM AS STRING*32 header
  DIM AS STRING filepath

  filepath=GETFILEPATH(file)

  i=INSTRREV(filepath, ":")
  ' #note: / #todo:
  ' a path containing ":" like "https://", "/C:/" will cause trouble, if
  ' the datafolder isn't specified correctly: "...<datafolder>:...[/...[/...]]"
  ' some might use the datafolder library unapparently through other libraries
  ' - at least a path like "C:\" should be translated into unix-like "/C/" or
  ' "/volume/C/" for instance. Common protocols have to be treated specially
  ' anyway
  ' #todo: network / HTTP and HTTPS support

  IF i=-1 OR DFISAFOLDER(filepath) THEN ' normal, direct file access
    ' "correcting" the path
    IF i<>-1 THEN
      i-=1
      filepath[i]=47 ' '/'
    ENDIF

    ' loading file in one go
    fn=FREEFILE()
    IF OPEN (filepath FOR BINARY AS #fn)=0 THEN
      l=LOF(fn)
      fbuffer=SPACE(l)
      GET #fn, 1, fbuffer
      CLOSE #fn
      RETURN TRUE
    ENDIF
  ELSE ' accessing an archive file
    archive_name=LEFT(filepath,i-1)
    filepath=MID(filepath, i+1)
    ' determining archive format by "header" / first 32 bytes
    fn=FREEFILE()
    IF OPEN(archive_name FOR BINARY AS #fn)=0 THEN
      GET #fn, , header
      CLOSE #fn
      IF LEFT(header,4)=CHR(&h50, &h4B, 3, 4) THEN
        ' ZIP archive
        ' #todo: unzip "file" in archive
        ' => ELSEIF other archive formats
      ELSE
        ' #todo: unsupported file format, let somehow know
      ENDIF
    ENDIF
  ENDIF
  RETURN FALSE ' error
END FUNCTION


FUNCTION DFLOAD(BYVAL file AS STRING, BYREF fbuffer AS UBYTE PTR, BYVAL encryption AS STRING="") AS BOOLEAN
  DIM AS INTEGER i, l, fn
  DIM AS STRING archive_name
  DIM AS STRING*32 header
  DIM AS STRING filepath

  filepath=GETFILEPATH(file)

  ' if fbuffer isn't 0, assumes already allocated buffer
  IF fbuffer THEN DEALLOCATE fbuffer

  i=INSTRREV(filepath, ":")
  ' #note: / #todo:
  ' paths containing ":" like "https://", "/C:/"
  ' #todo: network / HTTP and HTTPS support

  IF i=0 OR DFISAFOLDER(filepath) THEN ' normal, direct file access
    ' "correcting" the path
    IF i THEN
      i-=1
      filepath[i]=47 ' '/'
    ENDIF

    ' loading file in one go
    fn=FREEFILE()
    IF OPEN (filepath FOR BINARY AS #fn)=0 THEN
      l=LOF(fn)
      fbuffer=ALLOCATE(l)
      GET #fn, , *fbuffer, l
      CLOSE #fn
      RETURN TRUE
    ENDIF
  ELSE ' accessing an archive file
    archive_name=LEFT(filepath,i-1)
    filepath=MID(filepath, i+1)
    ' determining archive format by "header" / first 32 bytes
    fn=FREEFILE()
    IF OPEN(archive_name FOR BINARY AS #fn)=0 THEN
      GET #fn, , header
      CLOSE #fn
      IF LEFT(header,4)=CHR(&h50, &h4B, 3, 4) THEN
        ' ZIP archive
        ' #todo: unzip "file" in archive
        ' => ELSEIF other archive formats
      ELSE
        ' #todo: unsupported file format, let somehow know
      ENDIF
    ENDIF
  ENDIF
  RETURN FALSE ' error
END FUNCTION


FUNCTION DFLOADIMAGE(BYVAL file AS STRING, BYREF dest_buffer AS FB.IMAGE PTR, BYVAL encryption AS STRING="", BYREF palette_ AS ULONG PTR=0) AS BOOLEAN
  DIM AS LONG i, j, l, fn, w, h, bpp
  DIM AS STRING archive_name
  DIM AS STRING*32 header
  DIM AS STRING filepath
  DIM AS STRING extension

  filepath=GETFILEPATH(file)

  i=INSTRREV(filepath, ":")
  ' #note: / #todo:
  ' paths containing ":" like "https://", maybe "/C:/"
  ' #todo: network / HTTP and HTTPS support

  IF i=0 OR DFISAFOLDER(filepath) THEN ' normal, direct file access
    ' "correcting" the path
    IF i THEN
      i-=1
      filepath[i]=47 ' : to /
    ENDIF

    ' determine file type / format
    extension=RIGHT(filepath,4)
    extension=UCASE(extension)

    SELECT CASE extension
      CASE ".BMP"
        ' Windows(OS/2) bitmap file
        ' determining bitmap size and allocating buffer
        ' open file, determine dest. size
        fn=FREEFILE()
        IF OPEN (filepath FOR BINARY AS #fn)=0 THEN
          ' #note: uglyly GET's seek position is 1 higher than the file offset
          '        further, you can't load just 2 bytes into a 32 bit INTEGER
          ' check BMP head (16 bit)
          GET #fn, 1, h
          h=LITTLEENDIAN(CAST(INTEGER,h), 32) ' necessary for Big Endian systems
          h AND=&hFFFF
          IF h<>19778 THEN ' "BM"
            ' error, not a valid BMP file
            ' #todo:
            CLOSE #fn
            RETURN FALSE
          ENDIF
          ' check BMP compression
          GET #fn, 31, h
          h=LITTLEENDIAN(CAST(INTEGER,h), 32) ' necessary for Big Endian systems
          IF h THEN
            ' error, BMP compression => not supported
            ' #todo:
            CLOSE #fn
            RETURN FALSE
          ENDIF

          DIM AS UBYTE PTR destptr
          DIM AS INTEGER dummy, bmplwidth, destlwidth
          ' get bitmap dimensions - width height bits-per-pixel
          GET #fn, 19, w
          w=LITTLEENDIAN(CAST(INTEGER,w), 32) ' necessary for Big Endian systems
          GET #fn, 23, h
          h=LITTLEENDIAN(CAST(INTEGER,h), 32) ' necessary for Big Endian systems
          GET #fn, 29, bpp
          bpp=LITTLEENDIAN(CAST(INTEGER,bpp), 32) ' necessary for Big Endian systems
          IF w>0 AND h<>0 AND bpp<33 THEN
            IF bpp=1 OR bpp=4 OR bpp=15 OR (bpp AND 7)=0 THEN
              ' supported bpp values - 1, 4, 15, 8, 16, 24, 32
            ELSE
              ' illegal bpp value #todo: invalid file error
              RETURN FALSE
            ENDIF
          ELSE
            'illegal value #todo: invalid file error
            RETURN FALSE
          ENDIF
          ' #todo: maybe setting a limit for bitmap dimensions like 32768 pixels ?
          ' check destination buffer / allocate
          IF dest_buffer=0 THEN
            IF bpp<9 THEN
              dest_buffer=IMAGECREATE(w, ABS(h), 0, 8)
            ELSEIF bpp=15 OR bpp=16 THEN
              dest_buffer=IMAGECREATE(w, ABS(h), 0, 16)
            ELSEIF bpp=24 OR bpp=32 THEN
              dest_buffer=IMAGECREATE(w, ABS(h), 0, 32)
            ENDIF
          ELSE
            DIM AS INTEGER dw, dh, db
            IMAGEINFO dest_buffer, dw, dh, db
            ' "bpp" in IMAGEINFO() is BYTES per pixel instead of bits per pixel
            db SHL=3
            IF dw<>w OR dh<>ABS(h) OR db<bpp THEN
              ' not matching => maybe too small, destroy and create new
              IMAGEDESTROY dest_buffer
              IF bpp<9 THEN
                dest_buffer=IMAGECREATE(w, ABS(h), 0, 8)
              ELSEIF bpp=15 OR bpp=16 THEN
                dest_buffer=IMAGECREATE(w, ABS(h), 0, 16)
              ELSEIF bpp=24 OR bpp=32 THEN
                dest_buffer=IMAGECREATE(w, ABS(h), 0, 32)
              ENDIF
            ENDIF
          ENDIF
          IF dest_buffer=0 THEN RETURN FALSE ' #todo: error handling
          
          IF bpp<8 THEN
            ' maybe using "planes", normally 16 bit=16 pixels broad, 1 plane for each bpp
            ' => 4 bpp = 4 planes, 64 bits/8 bytes per 16 pixels
            ' #todo:
          ELSEIF bpp=8 THEN
            ' 1 byte per pixel
            bmplwidth=w
            destlwidth=w
          ELSEIF bpp=15 OR bpp=16 THEN
            ' 2 bytes per pixel
            bmplwidth=w*2
            destlwidth=w*2
          ELSEIF bpp=24 THEN
            ' 3 bytes per pixel
            bmplwidth=w*3
            destlwidth=w SHL 2
          ELSEIF bpp=32 THEN
            ' 4 bytes per pixel
            bmplwidth=w SHL 2
            destlwidth=w SHL 2
          ENDIF
          ' align BMP line width to a multiple of 4
          IF (bmplwidth AND 3) THEN ' not multiple of 4 => fill
            bmplwidth+=3
            bmplwidth AND=-4
          ENDIF
          ' align IMAGE line width to a multiple of 16
          IF (destlwidth AND 15) THEN ' not multiple of 16 => fill
            destlwidth+=15
            destlwidth AND=-16
          ENDIF

          IMAGEINFO dest_buffer, dummy, dummy, dummy, destlwidth, destptr

          ' read bitmap data
          GET #fn, 11, w ' offset to data
          w=LITTLEENDIAN(CAST(INTEGER,w), 32) ' necessary for Big Endian systems
          w+=1
          GET #fn, 35, l ' size of bitmap data
          l=LITTLEENDIAN(CAST(INTEGER,l), 32) ' necessary for Big Endian systems

          IF h>0 THEN
            ' bitmap data is upside-down
            destptr+=destlwidth*h
            i=h
            WHILE i
              i-=1
              destptr-=destlwidth
              GET #fn, w, *destptr, bmplwidth
              w+=bmplwidth
            WEND
          ELSE
            ' bitmap data is upright
            i=ABS(h)
            WHILE i
              i-=1
              GET #fn, w, *destptr, bmplwidth
              w+=bmplwidth
              destptr+=destlwidth
            WEND
          ENDIF
          ' read palette (if pointer is set)
          IF palette_ THEN
            GET #fn, 47, w
            w=LITTLEENDIAN(CAST(INTEGER,w), 32) ' necessary for Big Endian systems
            IF w>0 AND w<257 THEN
              ' #note: because it's a ULONG PTR, loads 4 times more bytes
              GET #fn, 55, *palette_, w
            ENDIF
          ENDIF
          CLOSE #fn
          ' correcting pixel data / adapting to FB.IMAGE format
          IF bpp=15 THEN
            ' #todo: ?
          ELSEIF bpp=24 THEN
            ' 24 bpp => 32 bpp / BB GG RR... => BB GG RR AA...
            IMAGEINFO dest_buffer, dummy, dummy, dummy, destlwidth, destptr
            w=dest_buffer->width
            h=dest_buffer->height
            WHILE h
              h-=1
              i=destlwidth*h
              j=i+(w*3)
              i+=w SHL 2 ' +w*4
              l=w-1 ' pixel count(er)
              DO
                i-=1
                destptr[i]=&hFF ' alpha
                IF l=0 THEN EXIT DO
                i-=1
                j-=1
                destptr[i]=destptr[j] ' red
                i-=1
                j-=1
                destptr[i]=destptr[j] ' green
                i-=1
                j-=1
                destptr[i]=destptr[j] ' blue
                l-=1
              LOOP
            WEND
          ENDIF
          RETURN TRUE
        ENDIF
      CASE ELSE
        ' #todo: FBImage library, maybe limit to supported file extensions
    END SELECT
  ELSE ' accessing an archive file
    archive_name=LEFT(filepath,i-1)
    filepath=MID(filepath, i+1)
    ' determining archive format by "header" / first 32 bytes
    fn=FREEFILE()
    IF OPEN(archive_name FOR BINARY AS #fn)=0 THEN
      GET #fn, , header
      CLOSE #fn
      IF LEFT(header,4)=CHR(&h50, &h4B, 3, 4) THEN
        ' ZIP archive
        ' #todo: unzip "file" in archive
        ' => ELSEIF other archive formats
      ELSE
        ' #todo: unsupported file format, let somehow know
      ENDIF
    ENDIF
  ENDIF
  RETURN FALSE ' error
END FUNCTION
