'
' fbext.bi
'
' FreeBASIC Extensions
'
' functions/subs that might become once / should be part of FreeBASIC core instruction set
'
' LITTLEENDIAN(INTEGER value, INTEGER bits)
' BIGENDIAN(INTEGER value, INTEGER bits)
'   converting/turning the given value if necessary. LITTLEENDIAN does nothing
'   on a machine working with Little Endian, on a Big Endian machine, it will
'   mirror the given value, BIGENDIAN is the counterpart
'   memory hook: "LITTLEENDIAN()" to take a Little Endian value correctly
'
' STRREPLACE(STRING destination, STRING search_string, STRING replace_string)
'   replaces in "destination" all occurrences of search_string with replace_string
'
' STRCOUNT(STRING source, STRING search_string)
'   returns count of "search_string" contained in "source"
'
' STRGETLISTITEM(STRING list, INTEGER index[, STRING divider])
'   returns a substring of a string list with the form "xx,xx,xx,xx..."
'   (or another specified divider)
'   index starts with 0 = first substring
'   Divider must be a single ASCII character except a nullbyte, " and \.
'   Strips off spaces except if using quotation marks to enclose the entries
'   => also ""xx","xx"...", further supporting the "escape sequences" \" and \\.
'   Dividers between quotation marks are ignored.
'
' STRGETLISTITEMCOUNT(STRING list[, STRING divider])
'   returns count of entries in a list with the form "xx,xx,xx,xx...", also
'   works with "<name>=<value>,<name>=<value>...", divider must not be quotation mark,
'   a divider between two quotation marks won't be counted
'
' STRSETLISTITEM(STRING list, [INTEGER index,][STRING divider,] STRING content)
'   sets a substring in a list of the form "xx,xx,xx,xx...", if index is out of
'   bounds or not specified, appends new entry
'
' STRGETLISTVALUE(STRING list, STRING name[, STRING divider])
'   returns a substring of a string list with the form "<name>=<value>,<name>=<value>"...
'   Addresses the substring by its "name" rather than its index
'   Quotation marks are counting as name and value limiters, so they have to be
'   masked with \"
'   Example: min=10.2,weather=bright equals "min"="10.2","weather"="bright"
'   Names must only consist of A-Z, a-z, 0-9, (, ), [, ], ., - and _
'   Dividers inside quotation mark sections are ignored
'   Using quotation marks can prevent parse errors, but you better may just
'   use e.g. CHR(8) as divider
'   Divider must be a single char and not nullbyte, ", =, . or \
'
' STRSETLISTVALUE(STRING list, STRING name, [STRING divider], STRING content)
'   sets a value in a list of the form "<name>=<value>,<name>=<value>"
'   won't add nor remove quotation marks, appends new entry if not existing
'
' --- "flags variable" --->
'   = BOOLEAN array, but each field is just one bit broad, mostly for (compact) storing
'   using e.g. "PUT/GET #.. variable" will store/load a sequence of bits in a file
'   - almost kind of cryptification, not as easy to read like "100101011..."
'   although 8 times more compact, in principle superflues memory saving cause there is
'   no use for billions of flags in one application
'   allows currently up to 524.288 flags
'
' FLAGSNEW(UBYTE PTR variable, INTEGER count)
'   creates a "flags variable" = BOOLEAN array
' DIMFLAGS(UBYTE PTR variable, INTEGER count)
'   like FLAGSNEW but can "re-dim" the array
'   #note: a new/undefined variable (pointer) must be 0
'
' FLAGSDELETE(UBYTE PTR variable)
'   deletes given flags variable / frees memory
'
' FLAGSSIZEOF(UBYTE PTR variable)
'   returns size of flags array, times 8 is max. count of flags
'
' GETFLAG(UBYTE PTR variable, INTEGER number)
'   returns TRUE or FALSE whether specified flag is set
'   index always starts with 0
' SETFLAG(UBYTE PTR variable, INTEGER number[, BOOLEAN value=TRUE])
'   sets specified flag
' CLRFLAG(UBYTE PTR variable, INTEGER number)
'   short for "SETFLAG( ., ., FALSE)"
' TGLFLAG(UBYTE PTR variable, INTEGER number)
'   short for "SETFLAG( ., ., NOT GETFLAG( ., .))"
' <--- "flags variable" ---
'
' CLASS("UDT") DATETIME:
'   Calculations between Unix (resp. Universal/UTC) time and Gregorian calendar days,
'   minutes and so on - Greenwich Mean Time.
'   Different time zones have to be determined separately because it depends on many
'   political/historical factors for each nation, especially the "summer time".
'   Use/set timezoneoffset (in seconds) to get the proper local time and date
'   #todo: maybe support of leap seconds => Universal time / UTC - requires tables
'   = LONGINT unixtime - calculates date/time values from given Unix time
'   = DATETIME pointintime - calculates Unix time from year, month, day, hour, minute, second
'   .GETUNIXTIME() - gets Unix time
'   .GETDATETIME()
' #todo: similar classes like DATETIME_JULIAN will support other calendar systems,
'        conversions should be based upon Unix time; there's no guarantee that a different
'        calendar system has days and months for instance
'

FUNCTION LITTLEENDIAN(BYREF value AS INTEGER, BYVAL bits AS INTEGER) AS INTEGER
  #IFDEF __FB_BIGENDIAN__
  ' Big Endian
  IF bits<9 THEN RETURN value ' nothing to mirror, joker

  IF bits<17 THEN
    ' mirror 2 bytes
    value=((value AND &hFF) SHL 8)+((value AND &hFF00) SHR 8)
  ELSEIF bits<33 THEN
    ' mirror 4 bytes
    value=((value AND &hFF) SHL 24)+((value AND &hFF00) SHL 8)+((value AND &hFF0000) SHR 8)+((value AND &hFF0000) SHR 24)
  ELSEIF bits<65 THEN
    ' mirror 8 bytes
    ' #todo:
  ELSEIF bits<129 THEN
    ' mirror 16 bytes
    ' #todo:
  ENDIF
  ' #todo: maybe also mirror 6, 12 bytes, maybe 3, 5, 7, 9, 10...?
  #ENDIF
  RETURN value    
END FUNCTION


FUNCTION BIGENDIAN(BYREF value AS INTEGER, BYVAL bits AS INTEGER) AS INTEGER
  #IFNDEF __FB_BIGENDIAN__
  ' Little Endian
  IF bits<9 THEN RETURN value ' nothing to mirror, joker

  IF bits<17 THEN
    ' mirror 2 bytes
    value=((value AND &hFF) SHL 8)+((value AND &hFF00) SHR 8)
  ELSEIF bits<33 THEN
    ' mirror 4 bytes
    value=((value AND &hFF) SHL 24)+((value AND &hFF00) SHL 8)+((value AND &hFF0000) SHR 8)+((value AND &hFF0000) SHR 24)
  ELSEIF bits<65 THEN
    ' mirror 8 bytes
    ' #todo:
  ELSEIF bits<129 THEN
    ' mirror 16 bytes
    ' #todo:
  ENDIF
  #ENDIF
  RETURN value    
END FUNCTION



FUNCTION STRREPLACE(BYREF destination AS STRING, BYVAL search_string AS STRING, BYVAL replace_string AS STRING) AS STRING
  DIM AS INTEGER i, l

  IF search_string="" THEN ' corrupt value
    destination=""
    RETURN ""
  ENDIF

  l=LEN(search_string)+1 ' #note: performance hack
  i=INSTR(destination, search_string)
  WHILE i
    i-=1 ' part of the hack
    destination=LEFT(destination, i)+replace_string+MID(destination, i+l)
    i=INSTR(i+LEN(replace_string), destination, search_string)
  WEND
  RETURN destination
END FUNCTION


FUNCTION STRCOUNT(BYVAL source AS STRING, BYVAL search_string AS STRING) AS INTEGER
  DIM AS INTEGER i, count

  count=0
  i=INSTR(source, search_string)
  WHILE i
    count+=1
    i+=LEN(search_string)
    i=INSTR(i, source, search_string)
  WEND

  RETURN count
END FUNCTION


FUNCTION STRGETLISTITEM(BYVAL list AS STRING, BYVAL index AS INTEGER, BYVAL divider AS STRING=",") AS STRING
  DIM AS STRING result
  DIM AS INTEGER i, l
  DIM AS UBYTE c, d

  ' validate divider
  IF LEN(divider)<>1 THEN RETURN "" ' more than 1 char is invalid => error
  d=divider[0]
  IF d=0 OR d>127 OR d=34 OR d=92 THEN RETURN "" ' invalid char => error

  ' "browse" the list
  i=0
  l=LEN(list)
  result=""
  WHILE i<l AND index>=0
    c=list[i]
    IF c=34 THEN ' quotation mark
      ' all between quotation marks is part of the "item"
      i+=1
      WHILE i<l
        c=list[i]
        IF c=34 THEN EXIT WHILE
        IF c=92 THEN ' backslash => "masking" following \ or "
          i+=1
          IF i>=l THEN EXIT WHILE
          c=list[i]
          IF index=0 THEN ' item index reached
            IF c<>92 AND c<>34 THEN result+="\" ' no effect in any other case
          ENDIF
        ENDIF
        IF index=0 THEN ' item index reached
          result+=CHR(c)
        ENDIF
        i+=1
      WEND
      WHILE i<l AND c<>d ' get next divider(comma)
        c=list[i]
        i+=1
      WEND
      index-=1 ' end of item => next item
    ELSEIF c=92 THEN ' backslash
      i+=1
      IF i>=l THEN EXIT WHILE
      c=list[i]
      IF index=0 THEN ' item index reached
        IF c<>92 AND c<>34 THEN result+="\" ' no effect in any other case
      ENDIF
    ELSEIF c<>d THEN ' not divider
      IF c<33 THEN ' "space chars" => ignore
        WHILE i<l
          c=list[i]
          IF c=d OR c>32 THEN EXIT WHILE
          i+=1
        WEND
      ELSE
        IF index=0 THEN ' item index reached
          result+=CHR(c)
        ENDIF
        i+=1
      endif
    ELSE ' divider reached
      index-=1
      i+=1
    ENDIF
  WEND

  RETURN result
END FUNCTION


FUNCTION STRGETLISTITEMCOUNT(BYVAL list AS STRING, BYVAL divider AS STRING=",") AS INTEGER
  DIM AS INTEGER i, counter, l
  DIM AS UBYTE c, d

  IF divider="" THEN RETURN 0 ' invalid value
  d=divider[0] ' only 1 char allowed

  counter=0
  l=LEN(list)
  i=0
  DO
    c=list[i]
    IF c=d THEN
      counter+=1
      i+=1
      CONTINUE DO
    ENDIF
    IF c=24 THEN ' quotation mark
      ' divider inside quoted string doesn't count
      ' jump to next (unmasked) quotation mark
      DO
        c=list[i]
        IF c=92 THEN ' backslash masks next char
          i+=2
          CONTINUE DO
        ELSEIF c=34 THEN ' quotation mark reached
          i+=1
          EXIT DO
        ELSEIF i>=l THEN ' end reached
          EXIT DO, DO
        ENDIF
        i+=1
      LOOP
    ENDIF
    i+=1
  LOOP

  RETURN counter
END FUNCTION


DECLARE FUNCTION STRSETLISTITEM OVERLOAD (BYVAL list AS STRING, BYVAL index AS INTEGER, BYVAL divider AS STRING=",", BYVAL content AS STRING) AS INTEGER
DECLARE FUNCTION STRSETLISTITEM (BYVAL list AS STRING, BYVAL divider AS STRING=",", BYVAL content AS STRING) AS INTEGER
FUNCTION STRSETLISTITEM(BYVAL list AS STRING, BYVAL content AS STRING) AS INTEGER
  ' wrapper
  RETURN STRSETLISTITEM(list, ",", content)
END FUNCTION


FUNCTION STRSETLISTITEM(BYVAL list AS STRING, BYVAL divider AS STRING=",", BYVAL content AS STRING) AS INTEGER
  ' just appending a new entry
  DIM AS INTEGER i
  DIM AS UBYTE c

  IF divider="" THEN RETURN 0 ' invalid value

  divider=LEFT(divider, 1) ' only 1 char allowed

  IF content>"" AND content[0]=34 THEN ' content seems to be a quoted string
    i=LEN(content)-1
    IF content[i]=34 THEN ' correctly quoted string
      content=LEFT(content, i) ' remove closing quotation mark
    ENDIF
    i=1
    WHILE i<LEN(content)
      c=content[i]
      IF c=34 OR c=92 THEN ' quotation mark or backslash => masking
        content=LEFT(content, i)+"\"+MID(content, i+1)
        i+=2
      ELSE
        i+=1
      ENDIF
    WEND
    content+=CHR(13) ' append closing quotation mark
  ELSE ' content is not a quoted string => masking ALL quotation marks
    i=0
    WHILE i<LEN(content)
      c=content[i]
      IF c=34 THEN ' quotation mark => masking
        content=LEFT(content, i)+"\"+MID(content, i+1)
        i+=2
      ELSE
        i+=1
      ENDIF
    WEND
  ENDIF

  list+=divider+content

  RETURN STRGETLISTITEMCOUNT(list, divider)
END FUNCTION


FUNCTION STRSETLISTITEM(BYVAL list AS STRING, BYVAL index AS INTEGER, BYVAL divider AS STRING=",", BYVAL content AS STRING) AS INTEGER
  ' #todo:
  RETURN index
END FUNCTION


FUNCTION STRGETLISTVALUE(BYVAL list AS STRING, BYVAL vname AS STRING, BYVAL divider AS STRING=",") AS STRING
  ' #todo:
  RETURN ""
END FUNCTION



SUB FLAGSDELETE(BYREF flagsvar AS UBYTE PTR)
  IF flagsvar THEN
    flagsvar-=2
    DEALLOCATE flagsvar
    flagsvar=0
  ENDIF
END SUB


FUNCTION FLAGSSIZEOF(flagsvar AS UBYTE PTR) AS INTEGER
  DIM AS INTEGER size
  IF flagsvar=0 THEN RETURN 0

  flagsvar-=2
  size=flagsvar[1]
  size SHL=8
  size +=flagsvar[0]
  size SHL=4 ' times 8

  RETURN size
END FUNCTION


FUNCTION DIMFLAGS(BYREF flagsvar AS UBYTE PTR, size AS INTEGER) AS UBYTE PTR
  IF size<2 OR size>524288 THEN
    ' error, illegal value, delete flagsvar precautionary
    IF flagsvar THEN FLAGSDELETE(flagsvar)
    RETURN 0
  ENDIF

  IF flagsvar THEN ' already existing/allocated => reallocate
    IF FLAGSSIZEOF(flagsvar)<>size THEN
      size/=8
      size+=3 ' "safe byte" plus size data
      flagsvar-=2 ' real address
      flagsvar=REALLOCATE(flagsvar, size)
      IF flagsvar THEN ' allocation
        ' hiding size information in buffer
        size-=2
        flagsvar[0]=size AND 255
        flagsvar[1]=(size SHR 8) AND 255
        flagsvar+=2 ' faked address
      ENDIF

    ENDIF
  ELSE
    size/=8
    size+=3 ' "safe byte" plus size data
    flagsvar=ALLOCATE(size)
    IF flagsvar THEN ' allocation
      ' hiding size information in buffer
      size-=2
      flagsvar[0]=size AND 255
      flagsvar[1]=(size SHR 8) AND 255
      flagsvar+=2 ' faked address
    ENDIF
  ENDIF

  RETURN flagsvar
END FUNCTION


FUNCTION FLAGSNEW(BYREF flagsvar AS UBYTE PTR, size AS INTEGER) AS UBYTE PTR
  IF size<2 OR size>524288 THEN
    ' error, illegal value, delete flagsvar precautionary
    IF flagsvar THEN FLAGSDELETE(flagsvar)
    RETURN 0
  ENDIF

  IF flagsvar THEN RETURN flagsvar ' already existing/allocated, error

  size/=8
  size+=3 ' "safe byte" plus size data
  flagsvar=ALLOCATE(size)
  IF flagsvar THEN ' allocation
    ' hiding size information in buffer
    size-=2
    flagsvar[0]=size AND 255
    flagsvar[1]=(size SHR 8) AND 255
    flagsvar+=2 ' faked address
  ENDIF

  RETURN flagsvar
END FUNCTION


FUNCTION GETFLAG(flagsvar AS UBYTE PTR, flagno AS INTEGER) AS BOOLEAN
  DIM AS BOOLEAN flag

  IF (flagsvar[flagno SHR 4] AND (1 SHL (flagno MOD 8))) THEN RETURN TRUE ELSE RETURN FALSE
END FUNCTION


SUB SETFLAG(flagsvar AS UBYTE PTR, flagno AS INTEGER, value AS BOOLEAN=TRUE)
  IF value THEN ' TRUE
    flagsvar[flagno SHR 4] OR=(1 SHL (flagno MOD 8))
  ELSE
    flagsvar[flagno SHR 4] AND=NOT (1 SHL (flagno MOD 8))
  ENDIF
END SUB


SUB CLRFLAG(flagsvar AS UBYTE PTR, flagno AS INTEGER)
  flagsvar[flagno SHR 4] AND=NOT (1 SHL (flagno MOD 8))
END SUB


FUNCTION TGLFLAG(flagsvar AS UBYTE PTR, flagno AS INTEGER) AS BOOLEAN
  DIM AS UBYTE c
  
  c=(1 SHL (flagno MOD 8)) ' the bit
  flagsvar[flagno SHR 4] XOR=c
  IF (flagsvar[flagno SHR 4] AND c)=c THEN
    RETURN TRUE ' bit set
  ENDIF
  RETURN FALSE ' bit cleared
END FUNCTION


TYPE DATETIME
  PUBLIC:
  unixtime AS LONGINT
  second AS UBYTE ' 0-60 (including leap seconds)
  minute AS UBYTE ' 0-59
  hour AS UBYTE ' 0-23
  wday AS UBYTE ' week day 0-6 0 = Monday
  mday AS UBYTE ' (Gregorian) day of month 1-31
  yday AS USHORT ' day of year 1-366
  week AS UBYTE ' calendar week 1-53 ISO 8601
  month AS UBYTE ' 1-12
  year AS SHORT ' #note: 0 is 1 BC, e.g. -1 2 BC
  timezoneoffset AS LONG ' -54000-+54000 (seconds)

  DECLARE CONSTRUCTOR
  DECLARE OPERATOR LET(BYVAL unixtime AS LONGINT) ' calculates all values out of given Unix time
  DECLARE OPERATOR LET(date_n_time AS DATETIME) ' turns given DATETIME values into Unix time
  DECLARE FUNCTION GETUNIXTIME() AS LONGINT ' returns unixtime, uses current date values
END TYPE


CONSTRUCTOR DATETIME
' initialisation without parameter
' "null" is 1.1.1970 0:00 (+0 "time zone")
unixtime=0
second=0
minute=0
hour=0
wday=3 ' Thursday
mday=1
yday=1
week=1
month=1
year=1970
timezoneoffset=0
END CONSTRUCTOR


OPERATOR DATETIME.LET (BYVAL utime AS LONGINT)
  DIM AS LONGINT i, o

  unixtime=utime
  utime+=timezoneoffset

  IF utime>=0 THEN
    wday=(INT(utime/86400)+3) MOD 7 ' not affected by leap years
  ELSE
    wday=(10+(INT(utime/86400) MOD 7)) MOD 7 ' not affected by leap years
  ENDIF

  IF utime>=0 THEN ' after 1969/12/31
    IF utime<94608000 THEN ' before 1972/12/31
      year=INT(utime/31536000)+1970
      o=utime MOD 31536000
    ELSE
      i=63072000 ' 1972/1/1 0:00
      year=1971
      DO
        year+=1
        o=i
        IF (year MOD 400) AND ((year MOD 4) OR (year MOD 100)=0) THEN
          i+=31536000 ' standard year
        ELSE
          i+=31622400 ' leap year
        ENDIF
      LOOP WHILE i<=utime
    ENDIF
  ELSE ' before 1970/1/1
    IF utime>=-31536000 THEN ' after 1968/12/31
      ' 1968 is already a leap year
      year=1969
      o=-31536000
    ELSE
      o=-31536000 ' 1969/1/1 0:00
      year=1969
      DO
        year-=1
        IF (year MOD 400) AND ((year MOD 4) OR (year MOD 100)=0) THEN
          o-=31536000 ' standard year
        ELSE
          o-=31622400 ' leap year
        ENDIF
      LOOP WHILE o>=utime
    ENDIF
  ENDIF

  yday=int((utime-o)/86400)+1 ' day of the year
  i=yday

  IF (year MOD 4)=0 AND ((year MOD 100) OR (year MOD 400)=0) THEN
    ' leap year
    IF yday=60 THEN ' special case Feburary 29th
      month=2
      mday=29
      i=0
    ELSE
      IF yday>60 THEN i-=1 ' after February, 29th
    ENDIF
  ENDIF

  IF i THEN ' not February 29th
    SELECT CASE i
      CASE IS <=31:
        month=1
        mday=yday
      CASE IS <=59:
        month=2
        mday=yday-31
      CASE IS <=90:
        month=3
        mday=i-59
      CASE IS <=120:
        month=4
        mday=i-90
      CASE IS <=151:
        month=5
        mday=i-120
      CASE IS <=181:
        month=6
        mday=i-151
      CASE IS <=212:
        month=7
        mday=i-181
      CASE IS <=243:
        month=8
        mday=i-212
      CASE IS <=273:
        month=9
        mday=i-243
      CASE IS <=304:
        month=10
        mday=i-273
      CASE IS <=334:
        month=11
        mday=i-304
      CASE ELSE:
        month=12
        mday=i-334
    END SELECT
  ENDIF

  utime-=o
  second=(utime MOD 60)
  minute=INT((utime MOD 3600)/60)
  hour=INT((utime MOD 86400)/3600)

  ' #todo: week= combine yday + wday
  ' first week of the year is first week with a Thursday
END OPERATOR


OPERATOR DATETIME.LET (BYREF dnt AS DATETIME)
  ' assigning "properties" of valid value range
  ' #note: does NOT completely validate, data might be nonsense like "February 31st"

  ' may be also January 1st 1970...
  unixtime=dnt.unixtime

  ' "null" values of all the other
  second=0
  minute=0
  hour=0
  wday=0
  mday=1
  month=1
  week=1
  year=0
  timezoneoffset=0

  IF dnt.second<61 THEN
    second=dnt.second
  ENDIF
  IF dnt.minute<60 THEN
    minute=dnt.minute
  ENDIF
  IF dnt.hour<24 THEN
    hour=dnt.hour
  ENDIF
  IF dnt.wday<7 THEN
    wday=dnt.wday
  ENDIF
  IF dnt.mday AND dnt.mday<32 THEN
    mday=dnt.mday
  ENDIF
  IF dnt.month AND dnt.month<13 THEN
    month=dnt.month
  ENDIF
  IF dnt.week AND dnt.week<54 THEN
    week=dnt.week
  ENDIF
  year=dnt.year
  IF ABS(dnt.timezoneoffset)<54000 THEN
    timezoneoffset=dnt.timezoneoffset
  ENDIF
END OPERATOR


FUNCTION DATETIME.GETUNIXTIME() AS LONGINT
  ' calculates and returns the unixtime value
  DIM AS LONGINT i

  ' member variable validation
  ' assigning doesn't check automatically

  ' year might be not more than 13.8 billion years ago... - but can't be
  ' => no limitiation, except it can only reach
  '   from about 32000 BC to 32000 AD anyway

  IF month<1 OR month>12 THEN ' can only be 1-12
    month=1
  ENDIF

  IF mday<1 OR mday>31 THEN mday=1 ' can only be 1-31
  IF mday>28 THEN
    IF month=2 THEN ' a February
      IF (year MOD 4) THEN mday=28 ' not a leap year
      ' #todo: this works only for Julian leap years, but it's rather strange
      '        why one should e.g. assign a 29th February in the year 1900...
    ELSEIF mday>30 THEN
      IF month<8 AND (month MOD 2)=0 THEN
        mday=30
      ELSEIF month>7 AND (month MOD 2) THEN
        mday=30
      ENDIF
    ENDIF
  ENDIF

  IF hour<0 OR hour>23 THEN hour=0 ' only 0-23

  IF minute<0 OR minute>59 THEN minute=0 ' only 0-59

  IF second<0 OR second>60 THEN second=0 ' only 0-60 (leap second)
  ' #todo: maybe determining if correct leap second if 60

  ' let's start the calculation
  unixtime=(year-1970)*31536000 ' standard years (leap days missing)
  IF year>2099 THEN
    ' #todo: calculate Gregorian leap years
  ELSEIF year<1901 THEN
    ' #todo: calculate Gregorian leap years
  ELSE
    ' simple, Julian leap year calculation
    IF year>1972 THEN
      unixtime+=INT((year-1969)/4)*86400
    ELSEIF year<1969 THEN
      unixtime-=INT((1972-year)/4)*86400
    ENDIF
  ENDIF

  ' add month
  IF month>1 THEN ' not January
    unixtime+=2678400 ' 31 days
    IF month>2 THEN ' after February
      ' determine whether leap year => 28 or 29 days
      IF (year MOD 4)=0 THEN
        IF (year MOD 100) OR (year MOD 400)=0 THEN
          unixtime+=2505600 ' +29 days
        ELSE
          unixtime+=2419200 ' +28 days
        ENDIF
      ELSE
        unixtime+=2419200 ' +28 days
      ENDIF
      IF month>3 THEN
        unixtime+=(month-3)*2592000 ' 30 days per month
        IF month<9 THEN
          unixtime+=INT((month-2)/2)*86400 ' +1-3 days (March,May,July)
        ELSE ' 31 day months
          unixtime+=INT((month-7)/2+3)*86400 ' +3+1-3 days
        ENDIF
      ENDIF
    ENDIF
  ENDIF

  ' add day
  unixtime+=(mday-1)*86400

  ' add time
  unixtime+=hour*3600
  unixtime+=minute*60
  unixtime+=second
  unixtime+=timezoneoffset

  ' #todo: determine day of year & week (yday / week)

  ' determine day of week (not affected by leap years)
  IF unixtime>=0 THEN
    wday=(INT(unixtime/86400)+3) MOD 7
  ELSE
    wday=(10+(INT(unixtime/86400) MOD 7)) MOD 7
  ENDIF

  RETURN unixtime
END FUNCTION
