'
' system.bi
'
' platform-independent FreeBASIC OS interfacing library
'
' "universal path": almost a Unix like form of filepaths, may start with a
'     %<key>% code, UTF-8 file names, separators are "/",
'     allows "file://" as "protocol". "C:\Windows" may become "/C/Windows/"
'     It's the form, file paths should have in code. Will be converted into
'     the platform-dependent path when used (not yet by OPEN directly)
'
' GETUSERLANGUAGE() - determine user's (or system's)
'     language and country settings, returns ISO-639+ISO-3166 divided
'     by an underscore, mainly 2 letters, but might be also 3 letters long
'     => ISO 639-1 or ISO 639-3 all in minuscules,
'     ISO 3166-1 ALPHA 2 preferred, but also ALPHA 3, ISO 3166-2 and
'     ISO 3166-3 possible, all in majuscules and divisions(3166-2) separated
'     by a "-" => "<languagecode>_<COUNTRYCODE>-<DIVISIONCODE>", the first
'     both 2 or 3 letters long, the last one 1 to 3 letters or numbers long
'
' GETUSERNAME() - get's "ID" of current user (its filepath name)
'
' GETFILEPATH(STRING source) - returns platform-dependent path of given
'     universal path. Won't remove the ":" of datafolders.
'     leading keys for special paths: have to stand at the beginning of
'     a file path, having the form: "%<key>%<folder>/<folder>/<file>"
'
'     %USERHOME%, %SHAREDHOME% - home folders for local user and all users
'     *SETTINGS, *DOCUMENTS, *DESKTOP, *DOWNLOADS, *GRAPHICS, *AUDIOS, *MEDIA,
'     *SAVEGAMES - all as USER- and SHARED- version
'     %USERSETTINGS% - place for config files, use a subfolder if for more than
'     one config file, e.g. using the name of PRJGETID()
'     %FBDATA% - FreeBASIC's data folder; non-application data, FreeBASIC "internal"
'     named "FreeBASIC.rsc" (sic! case-sensitive !)
'     %APPDATA% - place of current app's data (not config files),
'     searching first in local folder, then "user-local", then all-users/shared.
'     Assuming a data folder with ending ".rsc", will return ".rsc:" instead of
'     ".rsc/". Without extension assuming an ordinary folder and returning "/" at
'     the end.
'     note: The temporary files folder should be used via a specialised library
'     like the trash bin, and is therefore not regarded deliberately.
'     Something like the "Program Files" folder is only of use for something
'     like a setup wizard. As long as FreeBASIC won't be used for system tools...
'
' GETRETURNPATH(STRING source)
'     translates parts of a platform-dependent file path into the universal
'     format, mostly just exchanging e.g. backslashes with slashes
'     may also translate descriptive/localised folder names entered by the user
'     and "soft-links" into the real, system internal path
'
' NOTE: all file paths are returned with a '/' (or ':') at their end if
'       being a folder
'
DECLARE FUNCTION GETUSERNAME OVERLOAD AS STRING

#INCLUDE ONCE "prj.bi"

#IF Defined(__FB_WIN32__)
#INCLUDE "windows.bi"
#ENDIF

FUNCTION GETUSERLANGUAGE() AS STRING
  ' special thanks to 'jojo' and 'grindstone'
  DIM langcode AS STRING="en" ' default (& final fallback)
  DIM i AS INTEGER
  ' #todo: support of 3-letter-codes in case / "fra-CAN" shouldn't be illegal, see locale library for details

  #If Defined(__FB_WIN32__)

  Scope

  Dim sBuffer As STRING*32, nResult As INTEGER

  nResult = GetLocaleInfoA(&h400, &h59, STRPTR(sBuffer), 32)
  If nResult > 0 Then
    langcode = sBuffer + "_"
    nResult = GetLocaleInfoA(&h400, &h5a, STRPTR(sBuffer), 32)
    IF nResult > 0 Then langcode = langcode + sBuffer
  End If

  End Scope

  #ElseIf Defined(__FB_LINUX__)

  langcode=ENVIRON("LANG")
  IF langcode="" THEN
    Scope

    Dim ff As Integer = FreeFile
    Open Pipe "locale" For Input As ff

    Dim sRetVal As String

    Do Until EOF(ff)
      Line Input #ff, sRetVal
      If LCase(Left(sRetVal, 5)) = "lang=" Then
        langcode = LCase(Mid(sRetVal, 6))
        If InStr(langcode, ".") THEN ' strip off encoding
          langcode = Left(langcode, InStr(langcode, ".") - 1)
        ELSE
          langcode = sRetVal
        EndIf
        Exit Do
      EndIf
    Loop

    Close #ff

    End Scope
  ELSE
    i=INSTR(langcode, ".") ' might be something like "el_GR.UTF-8" => only language wanted
    IF i THEN langcode=MID(langcode, 1, i-1)
  ENDIF

  #Endif

  RETURN langcode
END FUNCTION


FUNCTION GETRETURNPATH(BYVAL source AS STRING) AS STRING
  ' well-forming returned path strings
  ' convert Windows "\" folder dividers into "/"
  ' in case also Mac OS ":" dividers (only in
  ' classic-Mac code => data folders !)
  DIM AS INTEGER i
  DIM AS STRING path=source

  IF source="" THEN RETURN ""

  i=INSTR(path, "\")
  WHILE i
    MID(path, i, 1)="/"
    i=INSTR(i, path, "\")
  WEND
  RETURN path
END FUNCTION


FUNCTION GETUSERNAME() AS STRING
  DIM AS STRING username=""
  #IF Defined(__FB_WIN32__)
  username=ENVIRON("USERNAME")
  #ELSEIF Defined(__FB_LINUX__)
  username=ENVIRON("USER")
  #ENDIF
  RETURN username
END FUNCTION


'FUNCTION USERPREFSPATH() AS STRING
'  DIM AS STRING path=""
'  #IF Defined(__FB_WIN32__)
'  path=ENVIRON("LOCALAPPDATA")
'  IF path="" THEN
'    ' #todo:(?)
'  ELSE
'    path+="/"+PRJGETID()+"/"
'  ENDIF
'  #ELSEIF Defined(__FB_LINUX__)
'  path=ENVIRON("HOME")
'  IF path="" THEN
'    ' #todo:(?)
'  ENDIF
'  path+="/.config/"
'  IF NOT FILEEXISTS(path) THEN
'    path=ENVIRON("HOME")+"/config"
'  ENDIF
'  #ENDIF
'  RETURN GETRETURNPATH(path)
'END FUNCTION
'
'
'FUNCTION SHAREDPREFSPATH() AS STRING
'  DIM AS STRING path=""
'  #IF Defined(__FB_WIN32__)
'  path=ENVIRON("PUBLIC")
'  IF path="" THEN
'    ' #todo:(?)
'  ELSE
'    path+="/"+PRJGETID()+"/"
'  ENDIF
'  #ELSEIF Defined(__FB_LINUX__)
'  path="/etc/"
'  IF path="" THEN
'    ' #todo:(?)
'  ENDIF
'  #ENDIF
'  RETURN GETRETURNPATH(path)
'END FUNCTION


FUNCTION GETSPECIALPATH(BYVAL key AS STRING) AS STRING
  ' internal helper function for GETFILEPATH()
  DIM AS STRING path

  #IF Defined(__FB_WIN32__)
  ' *** WINDOWS part ***
  ' #todo: still rather sloppy

  SELECT CASE key
    CASE "USERHOME"
      path=ENVIRON("HOMEPATH")
      IF path<>"" THEN
        path+="/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDHOME"
      path=ENVIRON("PUBLIC")
      IF path<>"" THEN
        path+="/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "USERSETTINGS"
      path=ENVIRON("LOCALAPPDATA")
      IF path<>"" THEN
        path+="/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDSETTINGS"
      ' #note: C:\Users\Public\AppData\Local\ would have made more sense, Microsoft...
      path=ENVIRON("APPDATA")
      IF path<>"" THEN
        path+="/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "USERDOCUMENTS"
      path=ENVIRON("HOMEPATH")
      IF path<>"" THEN
        path+="/Documents/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDDOCUMENTS"
      ' #todo: #toolazy
    CASE "USERDESKTOP"
      path=ENVIRON("HOMEPATH")
      IF path<>"" THEN
        path+="/Desktop/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDDESKTOP"
      ' #todo: #toolazy
    CASE "USERDOWNLOADS"
      path=ENVIRON("HOMEPATH")
      IF path<>"" THEN
        path+="/Downloads/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDDOWNLOADS"
      ' #todo: #toolazy
    CASE "USERGRAPHICS"
      path=ENVIRON("HOMEPATH")
      IF path<>"" THEN
        path+="/Pictures/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDGRAPHICS"
      ' #todo: #toolazy
    CASE "USERAUDIOS"
      path=ENVIRON("HOMEPATH")
      IF path<>"" THEN
        path+="/Music/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDAUDIOS"
      ' #todo: #toolazy
    CASE "USERMEDIA"
      path=ENVIRON("HOMEPATH")
      IF path<>"" THEN
        path+="/Videos/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDMEDIA"
      ' #todo: #toolazy
    CASE "USERSAVEGAMES"
      path=ENVIRON("HOMEPATH")
      IF path<>"" THEN
        path+="/Saved Games/"
        path=GETRETURNPATH(path)
      ENDIF
    CASE "SHAREDSAVEGAMES"
      ' #todo: #toolazy
    CASE "APPDATA"
      ' data folders always have a file extension, default is ".rsc"
      ' simple folders are returned with appended "/"

      ' first check local folder
      path=PRJGETID()
      IF DIR(path, 255)="" THEN
        path+=".rsc"
        IF DIR(path, 255)="" THEN path="" ELSE path+=":"
      ELSE
        path+="/"
      ENDIF

      IF path="" THEN
        ' local user app data path
        path=ENVIRON("LOCALAPPDATA") ' ...Users/.../AppData/Local/Myapp(.rsc)/
        IF path<>"" THEN
          path+="/"+PRJGETID()
          IF DIR(path, 255)="" THEN
            path+=".rsc"
            IF DIR(path, 255)="" THEN path="" ELSE path+=":"
          ELSE
            path+="/"
          ENDIF
        ENDIF
      ENDIF

      IF path="" THEN
        ' global app data path
        path=ENVIRON("ALLUSERSPROFILE") ' C:\ProgramData/Myapp(.rsc)/
        IF path<>"" THEN
          path+="/"+PRJGETID()
          IF DIR(path, 255)="" THEN
            path+=".rsc"
            IF DIR(path, 255)="" THEN path="" ELSE path+=":"
          ELSE
            path+="/"
          ENDIF
        ENDIF
      ELSE
        path+=":"
      ENDIF
      path=GETRETURNPATH(path)
    CASE "FBDATA"
      ' first check local folder
      path="FreeBASIC.rsc"
      IF DIR(path, 255)="" THEN path="" ELSE path+=":"

      IF path="" THEN
        ' local user app data path
        path=ENVIRON("LOCALAPPDATA") ' ...Users/.../AppData/Local/FreeBASIC.rsc/
        IF path<>"" THEN
          path+="/FreeBASIC.rsc"
          IF DIR(path, 255)="" THEN path="" ELSE path+=":"
        ENDIF
      ENDIF

      IF path="" THEN
        ' global app data path
        path=ENVIRON("ALLUSERSPROFILE") ' C:\ProgramData/FreeBASIC.rsc/
        IF path<>"" THEN
          path+="/FreeBASIC.rsc"
          IF DIR(path, 255)="" THEN path="" ELSE path+=":"
        ENDIF
      ENDIF
      path=GETRETURNPATH(path)
    CASE ELSE
      RETURN "" ' illegal => error
  END SELECT
  #ELSEIF Defined(__FB_LINUX__)
  ' *** LINUX part ***
  ' #todo: paths are to approve by the "high council of Linux"...
  '        idea: make paths to set by the user via enviroment variables
  '        => ENVIRON("USERSAVEGAMES") returned for %USERSAVEGAMES%
  '        or a similar, recommended mechanism
  ' recommendable would be full lowercase short folders to be "Unix like" e.g.:
  ' doc, desk, dnl, gra, audio, media, save - at least document, desktop, download,
  ' graphic, multimedia, savegame otherwise
  SELECT CASE key
    CASE "USERHOME"
      path=ENVIRON("HOME")
      path=GETRETURNPATH(path)
    CASE "SHAREDHOME"
      path="/home/group/public/"
    CASE "USERSETTINGS"
      ' config files shouldn't spam the home folder nor be invisible
      ' the (hidden) .config folder is the most spread Linux convention though
      path=ENVIRON("HOME")
      path=GETRETURNPATH(path)+"/.config/"
    CASE "SHAREDSETTINGS"
      path="/home/group/public/.config/"
    CASE "USERDOCUMENTS"
      ' not THAT Unix like, but established
      path=ENVIRON("HOME")
      path=GETRETURNPATH(path)+"/Documents/"
    CASE "SHAREDDOCUMENTS"
      path="/home/group/public/Documents/"
    CASE "USERDESKTOP"
      path=ENVIRON("HOME")
      path=GETRETURNPATH(path)+"/Desktop/"
    CASE "SHAREDDESKTOP"
      path="/home/group/public/Desktop/"
    CASE "USERDOWNLOADS"
      path=ENVIRON("HOME")
      path=GETRETURNPATH(path)+"/Downloads/"
    CASE "SHAREDDOWNLOADS"
      path="/home/group/public/Downloads/"
    CASE "USERGRAPHICS"
      path=ENVIRON("HOME")
      path=GETRETURNPATH(path)+"/Pictures/"
    CASE "SHAREDGRAPHICS"
      path="/home/group/public/Pictures/"
    CASE "USERAUDIOS"
      path=ENVIRON("HOME")
      path=GETRETURNPATH(path)+"/Music/"
    CASE "SHAREDAUDIOS"
      path="/home/group/public/Music/"
    CASE "USERMEDIA"
      path=ENVIRON("HOME")
      path=GETRETURNPATH(path)+"/Videos/"
    CASE "SHAREDMEDIA"
      path="/home/group/public/Videos/"
    CASE "USERSAVEGAMES"
      ' following the Ubuntu(?) scheme, although normally not existing
      path=ENVIRON("HOME")+"/Savegames/"
    CASE "SHAREDSAVEGAMES"
      path="/home/group/public/Savegames/"
    CASE "APPDATA"
      ' data folders always have a file extension, default is ".rsc"
      ' simple folders are returned with appended "/"

      ' first check local folder
      path=PRJGETID()
      IF DIR(path, 255)="" THEN
        path+=".rsc"
        IF DIR(path, 255)="" THEN path="" ELSE path+=":"
      ELSE
        path+="/"
      ENDIF

      ' global app data path
      path="/usr/share/"+PRJGETID()
      IF DIR(path, 255)="" THEN
        path+=".rsc"
        IF DIR(path, 255)="" THEN path="" ELSE path+=":"
      ELSE
        path+="/"
      ENDIF
    CASE "FBDATA"
      path="FreeBASIC.rsc"
      IF DIR(path, 255)="" THEN path="" ELSE path+=":"

      IF path="" THEN
        ' local user app data path
        path="/usr/share/FinalBASIC.rsc"
        IF DIR(path, 255)="" THEN path="" ELSE path+=":"
      ENDIF
    CASE ELSE
      RETURN "" ' illegal => error
  END SELECT
  #ENDIF

  RETURN path
END FUNCTION


FUNCTION GETFILEPATH(BYVAL source AS STRING) AS STRING
  DIM AS INTEGER i
  DIM AS STRING path

  IF source="" THEN RETURN ""

  IF source[0]=37 THEN ' starting with % => path keyword
    i=INSTR(2, source, "%")
    IF i<>0 THEN
      path=MID(source, 2, i-2)
      ' %USERHOME%, %SHAREDHOME%, %APPDATA% etc.
      path=GETSPECIALPATH(path)+MID(source, i+1)
      path=GETRETURNPATH(path)
    ENDIF
  ELSE ' nothing to do, except maybe syntax validation
    path=GETRETURNPATH(source)
  ENDIF

  ' if invalid path, path might be "" at this point
  RETURN path
END FUNCTION
